import sys
import os
import base64

# This tries to reassemble files from dblock and dindex files, without dlist.
# You don't really have any intact backup in this case, so results are rough.

# The dlist has the file names and information needed to do a proper restore.
# This tool should leave all versions of all files of all sizes -- plus more.
# Sorting through all that can be somewhat tool-assisted, but much is by you.

# If your files need decrypting, decrypt with AES Crypt or SharpAESCrypt.exe.
# You should have Windows Python, and enough free space for dblock and dindex
# folder and the file folder below that will hopefully be produced from that.
 
# If you need an external drive, avoid FAT type drive format, as it's slower.

# Before running script, you are responsible for setting up three subfolders:
list_folder = b"list"		# User fills this from the dindex files list.
block_folder = b"block"		# User fills this from the dblock files.
file_folder = b"file"		# User makes an empty folder to restore into.

# Example:
# for %i in (*.dindex.zip) do tar -xf %i list/* (ignore list not found error)
# mkdir block
# cd block
# for %i in (..\*.dblock.zip) do tar -xf %i (delete manifest, if you wish to)
# cd ..
# mkdir file
# Run the script by typing its name. It runs quietly, but inspect for errors.

# Upon success, reassembled files will be in file folder, leftovers in block.
# A file identifier such as "TriD -ae" could be used to try to guess content.

# Reassemble the multi-block files per the blocklists that list their blocks.
# Blocklists for deleted files might be around, but maybe without all blocks.
# Unlike Duplicati.CommandLine.RecoveryTool.exe, this discards partial files.
# These are kind of a "normal" thing, and the main goal is "noise" reduction.
print("Reassembling large multi-block files")
for blocklist_entry in os.scandir(list_folder):
	blocklist_name = blocklist_entry.name
	blocklist_file = open(os.path.join(list_folder, blocklist_name), "rb")
	assembled_file = open(os.path.join(file_folder, blocklist_name), "wb")
	try:
		while blockhash = blocklist_file.read(32):
			blockname = (base64.urlsafe_b64encode(blockhash))
			assembled_file.write(open(os.path.join(block_folder, blockname), "rb").read())
		assembled_file.close()
	except:
		assembled_file.close()
		os.remove(assembled_file.name)
	blocklist_file.close()

# If it stopped somewhere above, your list and block folders were not changed.
# Your file folder is a little suspect. If it finished, cleanups can commence.

# Reduce block clutter by deleting blocks that blocklists have put into files.
print("Cleaning up blocks for large files")
for blocklist_entry in os.scandir(list_folder):
	blocklist_name = blocklist_entry.name
	blocklist_file = open(os.path.join(list_folder, blocklist_name), "rb")
	while blockhash := blocklist_file.read(32):
		name = os.path.join(block_folder, base64.urlsafe_b64encode(blockhash))
		try:
			os.remove(name)
		except FileNotFoundError:
			pass
	blocklist_file.close()

# Reduce block clutter by deleting blocks that were already used as blocklist.
print("Cleaning up blocklist file blocks")
for blocklist_entry in os.scandir(list_folder):
	blocklist_name = blocklist_entry.name
	try:
		os.remove(os.path.join(block_folder, blocklist_name))
	except FileNotFoundError:
		pass

# Reduce block clutter by deleting blocks that look like metadata not content.
print("Cleaning up file metadata blocks")
for block_entry in os.scandir(block_folder):
	block_name = block_entry.name
	block_file = open(os.path.join(block_folder, block_name), "rb")
	start = block_file.read(26)
	block_file.close()
	if start == b'\xef\xbb\xbf{"win-ext:accessrules":':
		os.remove(os.path.join(block_folder, block_name))
	block_file.close()

# Before attempting any more cleanups, it's probably worth categorizing files.
# Multi-block files are in the file subfolder. The block folder has leftovers.

# TriD is sometimes confused by names starting with -, treating as parameters.
# One workaround seems to be to reference adding a folder prefix, e.g. file\*.

# TriD may also print out a disclaimer that it's best suited for binary files.
# You can probably use another tool to see if it's text, e.g. open in notepad,
# or make some assumptions, e.g. discard files containing unlikely characters.

# In the even more unfortunate case that there was a long-run backup that lost
# all its dlist files somehow, there may be blocks of files not in the backup.
# Ordinarily a compact would eventually run, but we're not in "ordinary" here.